(* $Id: smlng.ml,v 1.1 2001/07/28 15:02:37 ddr Exp $ *)
(* ICFP 2001 - Tonton Dada *)

open Printf;

(* Definitions *)

exception Lexer_error of string;

type token =
  [ Ttext of string
  | TBB | TEB
  | TBEM | TEEM
  | TBI | TEI
  | TBPL | TEPL
  | TBS | TES
  | TBTT | TETT
  | TBU | TEU
  | TB0 | TE0
  | TB1 | TE1
  | TB2 | TE2
  | TB3 | TE3
  | TB4 | TE4
  | TB5 | TE5
  | TB6 | TE6
  | TB7 | TE7
  | TB8 | TE8
  | TB9 | TE9
  | TBr | TEr
  | TBg | TEg
  | TBb | TEb
  | TBc | TEc
  | TBm | TEm
  | TBy | TEy
  | TBk | TEk
  | TBw | TEw ]
;

type doc =
  [ Dtext of string
  | Dtag of dtag and list doc ]
and dtag =
  [ SB
  | SEM
  | SI
  | SPL
  | SS
  | STT
  | SU
  | Ssize of int
  | Scolor of int ]
;

(* Buffer *)

value buff = ref (String.create 80);
value store len x =
  do {
    if len >= String.length buff.val then
      buff.val := buff.val ^ String.create (String.length buff.val)
    else ();
    buff.val.[len] := x;
    succ len
  }
;
value get_buff len = String.sub buff.val 0 len;

(* Lexer *)

value rec lexer =
  parser
  [ [: `'<'; t = less ? "tag expected"; `'>' ? "'>' expected" :] -> t
  | [: `x; t = text (store 0 x) :] -> Ttext t ]
and less =
  parser
  [ [: `'/'; t = tag True :] -> t
  | [: t = tag False :] -> t ]
and tag e =
  parser
  [ [: `'B' :] -> if e then TEB else TBB
  | [: `'E'; `'M' ? "'M' expected" :] -> if e then TEEM else TBEM
  | [: `'I' :] -> if e then TEI else TBI
  | [: `'P'; `'L' ? "'L' expected" :] -> if e then TEPL else TBPL
  | [: `'S' :] -> if e then TES else TBS
  | [: `'T'; `'T' ? "'T' expected" :] -> if e then TETT else TBTT
  | [: `'U' :] -> if e then TEU else TBU
  | [: `'0' :] -> if e then TE0 else TB0
  | [: `'1' :] -> if e then TE1 else TB1
  | [: `'2' :] -> if e then TE2 else TB2
  | [: `'3' :] -> if e then TE3 else TB3
  | [: `'4' :] -> if e then TE4 else TB4
  | [: `'5' :] -> if e then TE5 else TB5
  | [: `'6' :] -> if e then TE6 else TB6
  | [: `'7' :] -> if e then TE7 else TB7
  | [: `'8' :] -> if e then TE8 else TB8
  | [: `'9' :] -> if e then TE9 else TB9
  | [: `'r' :] -> if e then TEr else TBr
  | [: `'g' :] -> if e then TEg else TBg
  | [: `'b' :] -> if e then TEb else TBb
  | [: `'c' :] -> if e then TEc else TBc
  | [: `'m' :] -> if e then TEm else TBm
  | [: `'y' :] -> if e then TEy else TBy
  | [: `'k' :] -> if e then TEk else TBk
  | [: `'w' :] -> if e then TEw else TBw ]
and text len =
  parser
  [ [: `x when x <> '<' && x <> '>'; s :] -> text (store len x) s
  | [: :] -> get_buff len ]
;

(* Parser *)

value token_stream cs =
  Stream.from
    (fun i ->
       do {
         try Some (lexer cs) with
         [ Stream.Failure -> None
         | Stream.Error s -> raise (Lexer_error s) ]
       })
;

value rec dparser strm =
  List.rev (revd [] strm)
and revd gdl strm =
  match strm with parser
  [ [: `Ttext t :] -> revd [Dtext t :: gdl] strm
  | [: `TBB; k = dparser; `TEB ? "</B> expected" :] ->
      revd [Dtag SB k :: gdl] strm
  | [: `TBI; k = dparser; `TEI ? "</I> expected" :] ->
      revd [Dtag SI k :: gdl] strm
  | [: `TBEM; k = dparser; `TEEM ? "</EM> expected" :] ->
      revd [Dtag SEM k :: gdl] strm
  | [: `TBPL; k = dparser; `TEPL :] ->
      revd [Dtag SPL k :: gdl] strm
  | [: `TBS; k = dparser; `TES :] ->
      revd [Dtag SS k :: gdl] strm
  | [: `TBTT; k = dparser; `TETT :] ->
      revd [Dtag STT k :: gdl] strm
  | [: `TBU; k = dparser; `TEU :] ->
      revd [Dtag SU k :: gdl] strm
  | [: `TB0; k = dparser; `TE0 :] ->
      revd [Dtag (Ssize 0) k :: gdl] strm
  | [: `TB1; k = dparser; `TE1 :] ->
      revd [Dtag (Ssize 1) k :: gdl] strm
  | [: `TB2; k = dparser; `TE2 :] ->
      revd [Dtag (Ssize 2) k :: gdl] strm
  | [: `TB3; k = dparser; `TE3 :] ->
      revd [Dtag (Ssize 3) k :: gdl] strm
  | [: `TB4; k = dparser; `TE4 :] ->
      revd [Dtag (Ssize 4) k :: gdl] strm
  | [: `TB5; k = dparser; `TE5 :] ->
      revd [Dtag (Ssize 5) k :: gdl] strm
  | [: `TB6; k = dparser; `TE6 :] ->
      revd [Dtag (Ssize 6) k :: gdl] strm
  | [: `TB7; k = dparser; `TE7 :] ->
      revd [Dtag (Ssize 7) k :: gdl] strm
  | [: `TB8; k = dparser; `TE8 :] ->
      revd [Dtag (Ssize 8) k :: gdl] strm
  | [: `TB9; k = dparser; `TE9 :] ->
      revd [Dtag (Ssize 9) k :: gdl] strm
  | [: `TBr; k = dparser; `TEr :] ->
      revd [Dtag (Scolor 1) k :: gdl] strm
  | [: `TBg; k = dparser; `TEg :] ->
      revd [Dtag (Scolor 2) k :: gdl] strm
  | [: `TBb; k = dparser; `TEb :] ->
      revd [Dtag (Scolor 3) k :: gdl] strm
  | [: `TBc; k = dparser; `TEc :] ->
      revd [Dtag (Scolor 4) k :: gdl] strm
  | [: `TBm; k = dparser; `TEm :] ->
      revd [Dtag (Scolor 5) k :: gdl] strm
  | [: `TBy; k = dparser; `TEy :] ->
      revd [Dtag (Scolor 6) k :: gdl] strm
  | [: `TBk; k = dparser; `TEk :] ->
      revd [Dtag (Scolor 7) k :: gdl] strm
  | [: `TBw; k = dparser; `TEw :] ->
      revd [Dtag (Scolor 8) k :: gdl] strm
  | [: :] -> gdl ]
;

(* Optimization *)

type env = { env : int };

value flag_b  = 0b100000000000000;
value flag_em = 0b010000000000000;
value flag_i  = 0b001000000000000;
value flag_s  = 0b000100000000000;
value flag_tt = 0b000010000000000;
value bits_u      = 0b000001100000000;
value shift_u     = 8;
value bits_size   = 0b000000011110000;
value shift_size  = 4;
value bits_color  = 0b000000000001111;
value shift_color = 0;

value is_set flag {env = env} = env land flag <> 0;
value set flag {env = env} = {env = env lor flag};
value reset flag {env = env} = {env = env land (lnot flag)};
value invert flag {env = env} = {env = env lxor flag};

value set_size sz env =
  {env = env.env land (lnot bits_size) lor (sz lsl shift_size)}
;
value size_is_set sz env =
  (env.env land bits_size) lsr shift_size = sz
;
value set_color col env =
  {env = env.env land (lnot bits_color) lor (col lsl shift_color)}
;
value color_is_set col env =
  (env.env land bits_color) lsr shift_color = col
;

value strip_spaces s =
  loop 0 0 where rec loop i len =
    if i = String.length s then get_buff len
    else if i = 0 then loop (i + 1) (store 0 s.[0])
    else
      let len =
        match (s.[i-1], s.[i]) with
        [ ('\032' | '\013' | '\010' | '\009',
           '\032' | '\013' | '\010' | '\009') -> len
        | (_, c) -> store len c ]
      in
      loop (i + 1) len
;

value start_with_space env d =
  if is_set flag_tt env then False
  else
    loop d where rec loop =
      fun
      [ Dtext s ->
          if String.length s > 0 then
            match s.[0] with
            [ '\032' | '\013' | '\010' | '\009' -> True
            | _ -> False ]
          else False
      | Dtag t [d :: dl] -> t <> STT && loop d
      | _ -> False ]
;

value elim_ending_space d dl =
  let rec rebuild dl =
    fun
    [ Dtext s ->
        if String.length s = 1 then dl
        else
          let s = String.sub s 0 (String.length s - 1) in
          [Dtext s :: dl]
    | Dtag t dl1 ->
        [Dtag t (rebuild_list dl1) :: dl] ]
  and rebuild_list =
    fun
    [ [d] -> rebuild [] d
    | [d :: dl] -> [d :: rebuild_list dl]
    | [] -> [] ]
  in
  rebuild dl d
;

value end_with_space env d =
  if is_set flag_tt env then False
  else
    let rec ews =
      fun
      [ Dtext s ->
          if String.length s > 0 then
            match s.[String.length s - 1] with
            [ '\032' | '\013' | '\010' | '\009' -> True
            | _ -> False ]
          else False
      | Dtag t dl -> t <> STT && ews_list dl ]
    and ews_list =
      fun
      [ [d] -> ews d
      | [] -> False
      | [d :: dl] -> ews_list dl ]
    in
    ews d
;

value tag_len =
  fun
  [ SB | SI | Ssize _ | Scolor _ -> 1
  | SEM | STT -> 2
  | _ -> raise Not_found ]
;

value rec optim env dl =
  let rdl = revo env [] dl in
  loop [] rdl where rec loop dl =
    fun
    [ [d1; d2 :: rdl] ->
        let (dl, rdl) =
          try
            match (d1, d2) with
            [ (Dtag t2 [Dtag t3 k3 :: k2], Dtag t1 k1) when t1 = t3 ->
                let len_t2 = tag_len t2 in
                let (len_t1t3, tag_list, k1, k3) =
                  loop (tag_len t1) [t1] k1 k3
                  where rec loop len tag_list k1 k3 =
                    match (k1, k3) with
                    [ ([Dtag t1a k1a], [Dtag t3a k3a]) when t1a = t3a ->
                        loop (len + tag_len t1a + 5) [t1a :: tag_list] k1a k3a
                    | _ -> (len, tag_list, k1, k3) ]
                in
                if len_t2 < len_t1t3 then
                  let rdl =
                    loop (k1 @ [Dtag t2 k3]) tag_list where rec loop dl =
                      fun
                      [ [t] -> [Dtag t dl :: rdl]
                      | [t :: tl] -> loop [Dtag t dl] tl
                      | [] -> failwith "fail" ]
                  in
                  let dl = if k2 = [] then dl else [Dtag t2 k2 :: dl] in
                  (dl, rdl)
                else raise Not_found
(**)
            | (d1, d2) ->
                if end_with_space env d2 && start_with_space env d1 then
                  let dl = [d1 :: dl] in
                  let rdl = elim_ending_space d2 rdl in
                  (dl, rdl)
                else raise Not_found ]
(*
            | _ -> raise Not_found ]
*)
          with
          [ Not_found -> ([d1 :: dl], [d2 :: rdl]) ]
        in
        loop dl rdl
    | [d] -> [d :: dl]
    | [] -> dl ]
and revo env rdl = 
  fun
  [ [Dtext t :: gdl] ->
      match rdl with
      [ [Dtext t1 :: rdl] -> revo env rdl [Dtext (t1 ^ t) :: gdl]
      | _ ->
          let t = if is_set flag_tt env then t else strip_spaces t in
          revo env [Dtext t :: rdl] gdl ]
  | [Dtag SB k :: gdl] ->
      let is_set env k = is_set flag_b env in
      revo env (revo_tag SB k (set flag_b) is_set env rdl) gdl
  | [Dtag SI k :: gdl] ->
      match k with
      [ [Dtag SB k1] ->
          revo env rdl [Dtag SB [Dtag SI k1] :: gdl]
      | _ ->
          let is_set env k = is_set flag_i env in
          revo env (revo_tag SI k (set flag_i) is_set env rdl) gdl ]
  | [Dtag STT k :: gdl] ->
      match k with
      [ [Dtag (SB | SI as s) k1] ->
          revo env rdl [Dtag s [Dtag STT k1] :: gdl]
      | _ ->
          let is_set env k = is_set flag_tt env in
          revo env (revo_tag STT k (set flag_tt) is_set env rdl) gdl ]
  | [Dtag (Ssize sz) k :: gdl] ->
      match k with
      [ [Dtag (SB | SI | STT as s) k1] ->
          revo env rdl [Dtag s [Dtag (Ssize sz) k1] :: gdl]
      | _ ->
          let set = set_size sz in
          let is_set env k =
            if size_is_set sz env then True
            else
              match k with
              [ [Dtag (Ssize _) _] -> True
              | _ -> False ]
          in
          revo env (revo_tag (Ssize sz) k set is_set env rdl) gdl ]
  | [Dtag (Scolor col) k :: gdl] ->
      match k with
      [ [Dtag (SB | SI | STT | Ssize _ as s) k1] ->
          revo env rdl [Dtag s [Dtag (Scolor col) k1] :: gdl]
      | _ ->
          let set = set_color col in
          let is_set env k =
            if color_is_set col env then True
            else
              match k with
              [ [Dtag (Scolor _) _] -> True
              | _ -> False ]
          in
          revo env (revo_tag (Scolor col) k set is_set env rdl) gdl ]

  | [Dtag SEM k :: gdl] ->
      let k =
        let env = if is_set flag_s env then env else invert flag_em env in
        optim env k
      in
      let rdl =
        if is_set flag_s env then List.rev_append k rdl
        else
          match k with
          [ [Dtag SEM k] ->
              (* <EM><EM>txt</EM></EM> -> txt *)
              List.rev_append k rdl
          | [Dtag SEM k :: dl] ->
              (* <EM><EM>txt1</EM>txt2</EM> -> txt1<EM>txt2</EM> *)
              [Dtag SEM dl :: List.rev_append k rdl]
          | k ->
              match List.rev k with
              [ [Dtag SEM k :: dl] ->
                  (* <EM>txt1<EM>txt2</EM></EM> -> <EM>txt1</EM>txt2 *)
                  List.rev_append k [Dtag SEM (List.rev dl) :: rdl]
              | _ -> [Dtag SEM k :: rdl] ] ]
      in
      revo env rdl gdl
  | [Dtag SPL k :: gdl] ->
      let k =
        let env =
          reset
            (bits_u lor flag_s lor flag_em lor flag_i lor flag_b lor
             flag_tt)
            env
        in
        optim env k        
      in
      let rdl =
        if env.env land
             (bits_u lor flag_s lor flag_em lor flag_i lor flag_b lor
              flag_tt)
           = 0
        then
          (* <PL>...<PL>xxx</PL>...</PL> -> <PL>...xxx...</PL> *)
          List.rev_append k rdl
        else [Dtag SPL k :: rdl]
      in
      revo env rdl gdl
  | [Dtag SS k :: gdl] ->
      let k = optim (set flag_s (reset flag_em env)) k in
      revo env [Dtag SS k :: rdl] gdl
  | [Dtag SU k :: gdl] ->
      let rdl =
        let lev = (env.env land bits_u) lsr shift_u + 1 in
        let env =
          if lev > 3 then env
          else {env = env.env land (lnot bits_u) lor (lev lsl shift_u)}
        in
        let k = optim env k in
        if lev > 3 then
          (* <U><U><U><U>...</U></U></U></U> -> <U><U><U>...</U></U></U> *)
          List.rev_append k rdl
        else [Dtag SU k :: rdl]
      in
      revo env rdl gdl
(*
  | [Dtag x k :: dl] -> revo env [Dtag x (optim env k) :: rdl] dl
*)
  | [] -> rdl ]
and revo_tag tag k set is_set env rdl =
  let (k, rdl) =
    match rdl with
    [ [Dtag tag1 k1 :: rdl] when tag1 = tag -> (k1 @ k, rdl)
    | [(Dtext ("\032" | "\013" | "\010" | "\009") as d); Dtag tag1 k1
       :: rdl] when tag1 = tag -> (k1 @ [d :: k], rdl)
    | _ -> (k, rdl) ]
  in
  let k = optim (set env) k in
  if is_set env k then
    (* <B>...<B>txt</B>...</B> -> <B>...txt...</B> *)
    List.rev_append k rdl
  else [Dtag tag k :: rdl]
;

(* Printer *)

value color_of =
  fun
  [ 1 -> 'r'
  | 2 -> 'g'
  | 3 -> 'b'
  | 4 -> 'c'
  | 5 -> 'm'
  | 6 -> 'y'
  | 7 -> 'k'
  | 8 -> 'w'
  | _ -> 'r' ]
;

value rec print_doc oc =
  fun
  [ Dtext t -> fprintf oc "%s" t
  | Dtag tag k -> print_tag oc k tag ]
and print_tag oc k =
  fun
  [ SB -> fprintf oc "<B>%a</B>" print_doc_list k
  | SEM -> fprintf oc "<EM>%a</EM>" print_doc_list k
  | SI -> fprintf oc "<I>%a</I>" print_doc_list k
  | SPL -> fprintf oc "<PL>%a</PL>" print_doc_list k
  | SS -> fprintf oc "<S>%a</S>" print_doc_list k
  | STT -> fprintf oc "<TT>%a</TT>" print_doc_list k
  | SU -> fprintf oc "<U>%a</U>" print_doc_list k
  | Ssize x -> fprintf oc "<%d>%a</%d>" x print_doc_list k x
  | Scolor x ->
      let c = color_of x in
      fprintf oc "<%c>%a</%c>" c print_doc_list k c ]
and print_doc_list oc =
  fun
  [ [d :: dl] -> do { print_doc oc d; print_doc_list oc dl }
  | [] -> () ]
;

(* Launch *)

value main () =
  let cs = Stream.of_channel stdin in
  try
    let ts = token_stream cs in
    let dl = dparser ts in
    let dl = optim {env = 0} dl in
    do {
      print_doc_list stdout dl;
      flush stdout;
    }
  with e ->
    let ep = Stream.count cs in
    Stdpp.raise_with_loc (ep - 1, ep) e
;

try main () with e ->
  do {
    flush stdout;
    let e =
      match e with
      [ Stdpp.Exc_located (bp, ep) e -> do { eprintf "At location %d\n" ep; e }
      | e -> e ]
    in
    eprintf "Exception raised: %s\n" (Printexc.to_string e);
    flush stderr;
    exit 1;
  }
;
